module Task ( Task
            , newTask
            , addTask
            , addTasks
            , getTaskResults
            , ) where

import Control.Concurrent (forkIO)
import Control.Monad (replicateM, replicateM_, forever)
import Control.Concurrent.MVar
import Control.Concurrent.Chan
import Control.Exception

-- | submit actions to jobChan, worker will run it and put result in
-- resultChan.
worker :: Chan (IO a) -> Chan a -> IO ()
worker jobChan resultChan = forever $ do
  action <- readChan jobChan
  r <- action
  writeChan resultChan r

data Task a = Task {
      taskCount :: MVar Int
    , taskJobChan :: Chan (IO a)
    , taskResultChan :: Chan a }

data TaskException = TaskClosed deriving Show

instance Exception TaskException

-- | create a new task runner. TODO how to terminate workers when all job has
-- finished and task is closed?
newTask :: Int -> IO (Task a)
newTask n = do
  taskCountMVar <- newMVar 0
  chan1 <- newChan
  chan2 <- newChan
  let result = Task {
                 taskCount=taskCountMVar
               , taskJobChan=chan1
               , taskResultChan=chan2
               }
  replicateM_ n $ forkIO $ worker (taskJobChan result) (taskResultChan result)
  return result

-- | add action to task
addTask :: Task a -> IO a -> IO ()
addTask task action = do
  let tcount = taskCount task
  writeChan (taskJobChan task) action
  count <- takeMVar tcount
  putMVar tcount (count + 1)

-- | add a list of actions to task
addTasks :: Task a -> [IO a] -> IO ()
addTasks task actions = do
  let tcount = taskCount task
  writeList2Chan (taskJobChan task) actions
  count <- takeMVar tcount
  putMVar tcount (count + length actions)

-- | this is a blocking get. it will wait for all submitted tasks to finish
-- and return result. if new tasks are submitted, you can call getTaskResults
-- again.
getTaskResults :: Task a -> IO [a]
getTaskResults task = do
  n <- takeMVar (taskCount task)
  putMVar (taskCount task) 0
  replicateM n (readChan (taskResultChan task))
